from typing import Union

import sqlalchemy
import sqlalchemy.orm
import sqlalchemy.pool

from . import misc, application, group, auth, interface, job, base

ScopedSessionExt = Union[sqlalchemy.orm.Session, sqlalchemy.orm.scoping.ScopedSession]


class __ModelProxy:
    raw_session: ScopedSessionExt = None
    raw_engine = None

    def __getattr__(self, item):
        return getattr(self.raw_session, item)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            self.raw_session.rollback()
        self.raw_session.remove()


# noinspection PyTypeChecker
conn: ScopedSessionExt = __ModelProxy()


def init(url='sqlite://', engine_settings=None, connection_arguments=None):
    engine_defaults = dict(
        pool_use_lifo=True,
        pool_pre_ping=True,
        pool_recycle=360,
    )
    engine_settings = {**engine_defaults, **(engine_settings or {})}
    connection_arguments = connection_arguments or {}

    __ModelProxy.raw_engine = sqlalchemy.create_engine(
        url,
        poolclass=sqlalchemy.pool.QueuePool,
        **engine_settings,
        connect_args=connection_arguments,
    )
    __ModelProxy.raw_session = sqlalchemy.orm.scoped_session(sqlalchemy.orm.sessionmaker(bind=__ModelProxy.raw_engine))
    base.Base.metadata.create_all(__ModelProxy.raw_engine)



from typing import List

import riberry

log = riberry.log.make(__name__)
_executions_tracked = set()


class PoolExecutionTracker(riberry.app.backends.RiberryExecutionTracker):

    def _check_stale_executions(
            self,
            executions: List[riberry.model.job.JobExecution],
            app_instance: riberry.model.application.ApplicationInstance
    ):
        for execution in executions:
            if execution.task_id not in _executions_tracked:
                self._cancel_execution(execution=execution)

    def track_execution(self, root_id: str, app_instance: riberry.model.application.ApplicationInstance):
        log.debug(f'Tracking execution: root={root_id!r}')
        _executions_tracked.add(root_id)

    def _artifact_message(self, execution: riberry.model.job.JobExecution):
        return (
            f'The current executions\'s ID ({execution.task_id}) is no longer being tracked and has '
            f'therefore been cancelled. This usually occurs when the worker has terminated while a '
            f'task was either internally queued for execution or actively executing.'
        )
